@@ -5,7 +5,7 @@ This modules holds methods for generating predictions from a model.
 import os
 import sys
 from typing import List, Optional
-from logging import Logger
+import logging
 import numpy as np
 
 import torch
@@ -14,40 +14,44 @@ from torchtext.data import Dataset, Field
 from joeynmt.helpers import bpe_postprocess, load_config, make_logger,\
     get_latest_checkpoint, load_checkpoint, store_attention_plots
 from joeynmt.metrics import bleu, chrf, token_accuracy, sequence_accuracy
-from joeynmt.model import build_model, Model
+from joeynmt.model import build_model, Model, _DataParallel
+from joeynmt.search import run_batch
 from joeynmt.batch import Batch
 from joeynmt.data import load_data, make_data_iter, MonoDataset
 from joeynmt.constants import UNK_TOKEN, PAD_TOKEN, EOS_TOKEN
 from joeynmt.vocabulary import Vocabulary
 
+logger = logging.getLogger(__name__)
+
 
 # pylint: disable=too-many-arguments,too-many-locals,no-member
 def validate_on_data(model: Model, data: Dataset,
-                     logger: Logger,
                      batch_size: int,
                      use_cuda: bool, max_output_length: int,
                      level: str, eval_metric: Optional[str],
-                     loss_function: torch.nn.Module = None,
+                     n_gpu: int,
+                     compute_loss: bool = False,
                      beam_size: int = 1, beam_alpha: int = -1,
                      batch_type: str = "sentence",
-                     postprocess: bool = True
-                     ) \
+                     postprocess: bool = True,
+                     bpe_type: str = "subword-nmt",
+                     sacrebleu: dict = None) \
         -> (float, float, float, List[str], List[List[str]], List[str],
             List[str], List[List[str]], List[np.array]):
     """
     Generate translations for the given data.
-    If `loss_function` is not None and references are given,
+    If `compute_loss` is True and references are given,
     also compute the loss.
 
     :param model: model module
-    :param logger: logger
     :param data: dataset for validation
     :param batch_size: validation batch size
     :param use_cuda: if True, use CUDA
     :param max_output_length: maximum length for generated hypotheses
     :param level: segmentation level, one of "char", "bpe", "word"
     :param eval_metric: evaluation metric, e.g. "bleu"
-    :param loss_function: loss function that computes a scalar loss
+    :param n_gpu: number of GPUs
+    :param compute_loss: whether to computes a scalar loss
         for given inputs and targets
     :param beam_size: beam size for validation.
         If <2 then greedy decoding (default).
@@ -55,6 +59,8 @@ def validate_on_data(model: Model, data: Dataset,
         disabled if set to -1 (default).
     :param batch_type: validation batch type (sentence or token)
     :param postprocess: if True, remove BPE segmentation from translations
+    :param bpe_type: bpe type, one of {"subword-nmt", "sentencepiece"}
+    :param sacrebleu: sacrebleu options
 
     :return:
         - current_valid_score: current validation score [eval_metric],
@@ -67,6 +73,9 @@ def validate_on_data(model: Model, data: Dataset,
         - decoded_valid: raw validation hypotheses (before post-processing),
         - valid_attention_scores: attention scores for validation hypotheses
     """
+    assert batch_size >= n_gpu, "batch_size must be bigger than n_gpu."
+    if sacrebleu is None:   # assign default value
+        sacrebleu = {"remove_whitespace": True, "tokenize": "13a"}
     if batch_size > 1000 and batch_type == "sentence":
         logger.warning(
             "WARNING: Are you sure you meant to work on huge batches like "
@@ -92,20 +101,24 @@ def validate_on_data(model: Model, data: Dataset,
 
             batch = Batch(valid_batch, pad_index, use_cuda=use_cuda)
             # sort batch now by src length and keep track of order
-            sort_reverse_index = batch.sort_by_src_lengths()
+            sort_reverse_index = batch.sort_by_src_length()
 
             # run as during training with teacher forcing
-            if loss_function is not None and batch.trg is not None:
-                batch_loss = model.get_loss_for_batch(
-                    batch, loss_function=loss_function)
+            if compute_loss and batch.trg is not None:
+                batch_loss, _, _, _ = model(
+                    return_type="loss", src=batch.src, trg=batch.trg,
+                    trg_input=batch.trg_input, trg_mask=batch.trg_mask,
+                    src_mask=batch.src_mask, src_length=batch.src_length)
+                if n_gpu > 1:
+                    batch_loss = batch_loss.mean() # average on multi-gpu
                 total_loss += batch_loss
                 total_ntokens += batch.ntokens
                 total_nseqs += batch.nseqs
 
             # run as during inference to produce translations
-            output, attention_scores = model.run_batch(
-                batch=batch, beam_size=beam_size, beam_alpha=beam_alpha,
-                max_output_length=max_output_length)
+            output, attention_scores = run_batch(
+                model=model, batch=batch, beam_size=beam_size,
+                beam_alpha=beam_alpha, max_output_length=max_output_length)
 
             # sort outputs back to original order
             all_outputs.extend(output[sort_reverse_index])
@@ -115,7 +128,7 @@ def validate_on_data(model: Model, data: Dataset,
 
         assert len(all_outputs) == len(data)
 
-        if loss_function is not None and total_ntokens > 0:
+        if compute_loss and total_ntokens > 0:
             # total validation loss
             valid_loss = total_loss
             # exponent of token-level negative log prob
@@ -136,11 +149,12 @@ def validate_on_data(model: Model, data: Dataset,
 
         # post-process
         if level == "bpe" and postprocess:
-            valid_sources = [bpe_postprocess(s) for s in valid_sources]
-            valid_references = [bpe_postprocess(v)
+            valid_sources = [bpe_postprocess(s, bpe_type=bpe_type)
+                             for s in valid_sources]
+            valid_references = [bpe_postprocess(v, bpe_type=bpe_type)
                                 for v in valid_references]
-            valid_hypotheses = [bpe_postprocess(v) for
-                                v in valid_hypotheses]
+            valid_hypotheses = [bpe_postprocess(v, bpe_type=bpe_type)
+                                for v in valid_hypotheses]
 
         # if references are given, evaluate against them
         if valid_references:
@@ -149,12 +163,15 @@ def validate_on_data(model: Model, data: Dataset,
             current_valid_score = 0
             if eval_metric.lower() == 'bleu':
                 # this version does not use any tokenization
-                current_valid_score = bleu(valid_hypotheses, valid_references)
+                current_valid_score = bleu(
+                    valid_hypotheses, valid_references,
+                    tokenize=sacrebleu["tokenize"])
             elif eval_metric.lower() == 'chrf':
-                current_valid_score = chrf(valid_hypotheses, valid_references)
+                current_valid_score = chrf(valid_hypotheses, valid_references,
+                    remove_whitespace=sacrebleu["remove_whitespace"])
             elif eval_metric.lower() == 'token_accuracy':
-                current_valid_score = token_accuracy(
-                    valid_hypotheses, valid_references, level=level)
+                current_valid_score = token_accuracy(   # supply List[List[str]]
+                    list(decoded_valid), list(data.trg))
             elif eval_metric.lower() == 'sequence_accuracy':
                 current_valid_score = sequence_accuracy(
                     valid_hypotheses, valid_references)
@@ -166,12 +183,76 @@ def validate_on_data(model: Model, data: Dataset,
         decoded_valid, valid_attention_scores
 
 
+def parse_test_args(cfg, mode="test"):
+    """
+    parse test args
+    :param cfg: config object
+    :param mode: 'test' or 'translate'
+    :return:
+    """
+    if "test" not in cfg["data"].keys():
+        raise ValueError("Test data must be specified in config.")
+
+    batch_size = cfg["training"].get(
+        "eval_batch_size", cfg["training"].get("batch_size", 1))
+    batch_type = cfg["training"].get(
+        "eval_batch_type", cfg["training"].get("batch_type", "sentence"))
+    use_cuda = (cfg["training"].get("use_cuda", False)
+                and torch.cuda.is_available())
+    device = torch.device("cuda" if use_cuda else "cpu")
+    if mode == 'test':
+        n_gpu = torch.cuda.device_count() if use_cuda else 0
+        logger.info("Process device: %s, n_gpu: %d, batch_size per device: %d",
+            device, n_gpu, batch_size // n_gpu if n_gpu > 1 else batch_size)
+        eval_metric = cfg["training"]["eval_metric"]
+
+    elif mode == 'translate':
+        # in multi-gpu, batch_size must be bigger than n_gpu!
+        n_gpu = 1 if use_cuda else 0
+        logger.debug("Process device: %s, n_gpu: %d", device, n_gpu)
+        eval_metric = ""
+
+    level = cfg["data"]["level"]
+    max_output_length = cfg["training"].get("max_output_length", None)
+
+    # whether to use beam search for decoding, 0: greedy decoding
+    if "testing" in cfg.keys():
+        beam_size = cfg["testing"].get("beam_size", 1)
+        beam_alpha = cfg["testing"].get("alpha", -1)
+        postprocess = cfg["testing"].get("postprocess", True)
+        bpe_type = cfg["testing"].get("bpe_type", "subword-nmt")
+        sacrebleu = {"remove_whitespace": True, "tokenize": "13a"}
+        if "sacrebleu" in cfg["testing"].keys():
+            sacrebleu["remove_whitespace"] = cfg["testing"]["sacrebleu"] \
+                .get("remove_whitespace", True)
+            sacrebleu["tokenize"] = cfg["testing"]["sacrebleu"] \
+                .get("tokenize", "13a")
+
+    else:
+        beam_size = 1
+        beam_alpha = -1
+        postprocess = True
+        bpe_type = "subword-nmt"
+        sacrebleu = {"remove_whitespace": True, "tokenize": "13a"}
+
+    decoding_description = "Greedy decoding" if beam_size < 2 else \
+        "Beam search decoding with beam size = {} and alpha = {}". \
+            format(beam_size, beam_alpha)
+    tokenizer_info = f"[{sacrebleu['tokenize']}]" \
+        if eval_metric == "bleu" else ""
+
+    return batch_size, batch_type, use_cuda, n_gpu, level, \
+           eval_metric, max_output_length, beam_size, beam_alpha, \
+           postprocess, bpe_type, sacrebleu, decoding_description, \
+           tokenizer_info
+
+
 # pylint: disable-msg=logging-too-many-args
 def test(cfg_file,
          ckpt: str,
          output_path: str = None,
          save_attention: bool = False,
-         logger: Logger = None) -> None:
+         datasets: dict = None) -> None:
     """
     Main test function. Handles loading a model from checkpoint, generating
     translations and storing them and attention plots.
@@ -179,44 +260,39 @@ def test(cfg_file,
     :param cfg_file: path to configuration file
     :param ckpt: path to checkpoint to load
     :param output_path: path to output
+    :param datasets: datasets to predict
     :param save_attention: whether to save the computed attention weights
-    :param logger: log output to this logger (creates new logger if not set)
     """
 
-    if logger is None:
-        logger = make_logger()
-
     cfg = load_config(cfg_file)
+    model_dir = cfg["training"]["model_dir"]
 
-    if "test" not in cfg["data"].keys():
-        raise ValueError("Test data must be specified in config.")
+    if len(logger.handlers) == 0:
+        _ = make_logger(model_dir, mode="test")   # version string returned
 
     # when checkpoint is not specified, take latest (best) from model dir
     if ckpt is None:
-        model_dir = cfg["training"]["model_dir"]
         ckpt = get_latest_checkpoint(model_dir)
-        if ckpt is None:
-            raise FileNotFoundError("No checkpoint found in directory {}."
-                                    .format(model_dir))
         try:
             step = ckpt.split(model_dir+"/")[1].split(".ckpt")[0]
         except IndexError:
             step = "best"
 
-    batch_size = cfg["training"].get(
-        "eval_batch_size", cfg["training"]["batch_size"])
-    batch_type = cfg["training"].get(
-        "eval_batch_type", cfg["training"].get("batch_type", "sentence"))
-    use_cuda = cfg["training"].get("use_cuda", False)
-    level = cfg["data"]["level"]
-    eval_metric = cfg["training"]["eval_metric"]
-    max_output_length = cfg["training"].get("max_output_length", None)
-
     # load the data
-    _, dev_data, test_data, src_vocab, trg_vocab = load_data(
-        data_cfg=cfg["data"])
+    if datasets is None:
+        _, dev_data, test_data, src_vocab, trg_vocab = load_data(
+            data_cfg=cfg["data"], datasets=["dev", "test"])
+        data_to_predict = {"dev": dev_data, "test": test_data}
+    else:   # avoid to load data again
+        data_to_predict = {"dev": datasets["dev"], "test": datasets["test"]}
+        src_vocab = datasets["src_vocab"]
+        trg_vocab = datasets["trg_vocab"]
 
-    data_to_predict = {"dev": dev_data, "test": test_data}
+    # parse test args
+    batch_size, batch_type, use_cuda, n_gpu, level, eval_metric, \
+        max_output_length, beam_size, beam_alpha, postprocess, \
+        bpe_type, sacrebleu, decoding_description, tokenizer_info \
+        = parse_test_args(cfg, mode="test")
 
     # load model state from disk
     model_checkpoint = load_checkpoint(ckpt, use_cuda=use_cuda)
@@ -228,17 +304,16 @@ def test(cfg_file,
     if use_cuda:
         model.cuda()
 
-    # whether to use beam search for decoding, 0: greedy decoding
-    if "testing" in cfg.keys():
-        beam_size = cfg["testing"].get("beam_size", 1)
-        beam_alpha = cfg["testing"].get("alpha", -1)
-        postprocess = cfg["testing"].get("postprocess", True)
-    else:
-        beam_size = 1
-        beam_alpha = -1
-        postprocess = True
+    # multi-gpu eval
+    if n_gpu > 1 and not isinstance(model, torch.nn.DataParallel):
+        model = _DataParallel(model)
 
     for data_set_name, data_set in data_to_predict.items():
+        if data_set is None:
+            continue
+
+        dataset_file = cfg["data"][data_set_name] + "." + cfg["data"]["trg"]
+        logger.info("Decoding on %s set (%s)...", data_set_name, dataset_file)
 
         #pylint: disable=unused-variable
         score, loss, ppl, sources, sources_raw, references, hypotheses, \
@@ -246,16 +321,15 @@ def test(cfg_file,
             model, data=data_set, batch_size=batch_size,
             batch_type=batch_type, level=level,
             max_output_length=max_output_length, eval_metric=eval_metric,
-            use_cuda=use_cuda, loss_function=None, beam_size=beam_size,
-            beam_alpha=beam_alpha, logger=logger, postprocess=postprocess)
+            use_cuda=use_cuda, compute_loss=False, beam_size=beam_size,
+            beam_alpha=beam_alpha, postprocess=postprocess,
+            bpe_type=bpe_type, sacrebleu=sacrebleu, n_gpu=n_gpu)
         #pylint: enable=unused-variable
 
         if "trg" in data_set.fields:
-            decoding_description = "Greedy decoding" if beam_size < 2 else \
-                "Beam search decoding with beam size = {} and alpha = {}".\
-                    format(beam_size, beam_alpha)
-            logger.info("%4s %s: %6.2f [%s]",
-                        data_set_name, eval_metric, score, decoding_description)
+            logger.info("%4s %s%s: %6.2f [%s]",
+                        data_set_name, eval_metric, tokenizer_info,
+                        score, decoding_description)
         else:
             logger.info("No references given for %s -> no evaluation.",
                         data_set_name)
@@ -285,7 +359,7 @@ def test(cfg_file,
             logger.info("Translations saved to: %s", output_path_set)
 
 
-def translate(cfg_file, ckpt: str, output_path: str = None) -> None:
+def translate(cfg_file: str, ckpt: str, output_path: str = None) -> None:
     """
     Interactive translation function.
     Loads model from checkpoint and translates either the stdin input or
@@ -317,8 +391,6 @@ def translate(cfg_file, ckpt: str, output_path: str = None) -> None:
 
         return test_data
 
-    logger = make_logger()
-
     def _translate_data(test_data):
         """ Translates given dataset, using parameters from outer scope. """
         # pylint: disable=unused-variable
@@ -327,30 +399,24 @@ def translate(cfg_file, ckpt: str, output_path: str = None) -> None:
             model, data=test_data, batch_size=batch_size,
             batch_type=batch_type, level=level,
             max_output_length=max_output_length, eval_metric="",
-            use_cuda=use_cuda, loss_function=None, beam_size=beam_size,
-            beam_alpha=beam_alpha, logger=logger, postprocess=postprocess)
+            use_cuda=use_cuda, compute_loss=False, beam_size=beam_size,
+            beam_alpha=beam_alpha, postprocess=postprocess,
+            bpe_type=bpe_type, sacrebleu=sacrebleu, n_gpu=n_gpu)
         return hypotheses
 
     cfg = load_config(cfg_file)
+    model_dir = cfg["training"]["model_dir"]
+
+    _ = make_logger(model_dir, mode="translate")
+    # version string returned
 
     # when checkpoint is not specified, take oldest from model dir
     if ckpt is None:
-        model_dir = cfg["training"]["model_dir"]
         ckpt = get_latest_checkpoint(model_dir)
 
-    batch_size = cfg["training"].get(
-        "eval_batch_size", cfg["training"].get("batch_size", 1))
-    batch_type = cfg["training"].get(
-        "eval_batch_type", cfg["training"].get("batch_type", "sentence"))
-    use_cuda = cfg["training"].get("use_cuda", False)
-    level = cfg["data"]["level"]
-    max_output_length = cfg["training"].get("max_output_length", None)
-
     # read vocabs
-    src_vocab_file = cfg["data"].get(
-        "src_vocab", cfg["training"]["model_dir"] + "/src_vocab.txt")
-    trg_vocab_file = cfg["data"].get(
-        "trg_vocab", cfg["training"]["model_dir"] + "/trg_vocab.txt")
+    src_vocab_file = cfg["data"].get("src_vocab", model_dir + "/src_vocab.txt")
+    trg_vocab_file = cfg["data"].get("trg_vocab", model_dir + "/trg_vocab.txt")
     src_vocab = Vocabulary(file=src_vocab_file)
     trg_vocab = Vocabulary(file=trg_vocab_file)
 
@@ -367,6 +433,11 @@ def translate(cfg_file, ckpt: str, output_path: str = None) -> None:
                       include_lengths=True)
     src_field.vocab = src_vocab
 
+    # parse test args
+    batch_size, batch_type, use_cuda, n_gpu, level, _, \
+        max_output_length, beam_size, beam_alpha, postprocess, \
+        bpe_type, sacrebleu, _, _ = parse_test_args(cfg, mode="translate")
+
     # load model state from disk
     model_checkpoint = load_checkpoint(ckpt, use_cuda=use_cuda)
 
@@ -377,16 +448,6 @@ def translate(cfg_file, ckpt: str, output_path: str = None) -> None:
     if use_cuda:
         model.cuda()
 
-    # whether to use beam search for decoding, <2: greedy decoding
-    if "testing" in cfg.keys():
-        beam_size = cfg["testing"].get("beam_size", 1)
-        beam_alpha = cfg["testing"].get("alpha", -1)
-        postprocess = cfg["testing"].get("postprocess", True)
-    else:
-        beam_size = 1
-        beam_alpha = -1
-        postprocess = True
-
     if not sys.stdin.isatty():
         # input file given
         test_data = MonoDataset(path=sys.stdin, ext="", field=src_field)
