@@ -4,16 +4,15 @@ import torch.nn.functional as F
 from torch import Tensor
 import numpy as np
 
-from joeynmt.decoders import Decoder, TransformerDecoder
-from joeynmt.embeddings import Embeddings
+from joeynmt.decoders import TransformerDecoder
+from joeynmt.model import Model
+from joeynmt.batch import Batch
 from joeynmt.helpers import tile
 
-
-__all__ = ["greedy", "transformer_greedy", "beam_search"]
+__all__ = ["greedy", "transformer_greedy", "beam_search", "run_batch"]
 
 
-def greedy(src_mask: Tensor, embed: Embeddings, bos_index: int, eos_index: int,
-           max_output_length: int, decoder: Decoder,
+def greedy(src_mask: Tensor, max_output_length: int, model: Model,
            encoder_output: Tensor, encoder_hidden: Tensor)\
         -> (np.array, np.array):
     """
@@ -22,17 +21,14 @@ def greedy(src_mask: Tensor, embed: Embeddings, bos_index: int, eos_index: int,
     recurrent decoders and transformer_greedy for transformer decoders.
 
     :param src_mask: mask for source inputs, 0 for positions after </s>
-    :param embed: target embedding
-    :param bos_index: index of <s> in the vocabulary
-    :param eos_index: index of </s> in the vocabulary
     :param max_output_length: maximum length for the hypotheses
-    :param decoder: decoder to use for greedy decoding
+    :param model: model to use for greedy decoding
     :param encoder_output: encoder hidden states for attention
     :param encoder_hidden: encoder last state for decoder initialization
     :return:
     """
 
-    if isinstance(decoder, TransformerDecoder):
+    if isinstance(model.decoder, TransformerDecoder):
         # Transformer greedy decoding
         greedy_fun = transformer_greedy
     else:
@@ -40,30 +36,27 @@ def greedy(src_mask: Tensor, embed: Embeddings, bos_index: int, eos_index: int,
         greedy_fun = recurrent_greedy
 
     return greedy_fun(
-        src_mask, embed, bos_index, eos_index, max_output_length,
-        decoder, encoder_output, encoder_hidden)
+        src_mask, max_output_length, model, encoder_output, encoder_hidden)
 
 
 def recurrent_greedy(
-        src_mask: Tensor, embed: Embeddings, bos_index: int, eos_index: int,
-        max_output_length: int, decoder: Decoder,
+        src_mask: Tensor, max_output_length: int, model: Model,
         encoder_output: Tensor, encoder_hidden: Tensor) -> (np.array, np.array):
     """
     Greedy decoding: in each step, choose the word that gets highest score.
     Version for recurrent decoder.
 
     :param src_mask: mask for source inputs, 0 for positions after </s>
-    :param embed: target embedding
-    :param bos_index: index of <s> in the vocabulary
-    :param eos_index: index of </s> in the vocabulary
     :param max_output_length: maximum length for the hypotheses
-    :param decoder: decoder to use for greedy decoding
+    :param model: model to use for greedy decoding
     :param encoder_output: encoder hidden states for attention
     :param encoder_hidden: encoder last state for decoder initialization
     :return:
         - stacked_output: output hypotheses (2d array of indices),
         - stacked_attention_scores: attention scores (3d array)
     """
+    bos_index = model.bos_index
+    eos_index = model.eos_index
     batch_size = src_mask.size(0)
     prev_y = src_mask.new_full(size=[batch_size, 1], fill_value=bos_index,
                                dtype=torch.long)
@@ -76,22 +69,24 @@ def recurrent_greedy(
     # pylint: disable=unused-variable
     for t in range(max_output_length):
         # decode one single step
-        logits, hidden, att_probs, prev_att_vector = decoder(
-            encoder_output=encoder_output,
-            encoder_hidden=encoder_hidden,
-            src_mask=src_mask,
-            trg_embed=embed(prev_y),
-            hidden=hidden,
-            prev_att_vector=prev_att_vector,
-            unroll_steps=1)
-        # logits: batch x time=1 x vocab (logits)
+        with torch.no_grad():
+            logits, hidden, att_probs, prev_att_vector = model(
+                return_type="decode",
+                trg_input=prev_y,
+                encoder_output=encoder_output,
+                encoder_hidden=encoder_hidden,
+                src_mask=src_mask,
+                unroll_steps=1,
+                decoder_hidden=hidden,
+                att_vector=prev_att_vector)
+            # logits: batch x time=1 x vocab (logits)
 
         # greedy decoding: choose arg max over vocabulary in each step
         next_word = torch.argmax(logits, dim=-1)  # batch x time=1
         output.append(next_word.squeeze(1).detach().cpu().numpy())
         prev_y = next_word
         attention_scores.append(att_probs.squeeze(1).detach().cpu().numpy())
-        # batch, max_src_lengths
+        # batch, max_src_length
 
         # check if previous symbol was <eos>
         is_eos = torch.eq(next_word, eos_index)
@@ -107,27 +102,23 @@ def recurrent_greedy(
 
 # pylint: disable=unused-argument
 def transformer_greedy(
-        src_mask: Tensor, embed: Embeddings,
-        bos_index: int, eos_index: int,
-        max_output_length: int, decoder: Decoder,
+        src_mask: Tensor, max_output_length: int, model: Model,
         encoder_output: Tensor, encoder_hidden: Tensor) -> (np.array, None):
     """
     Special greedy function for transformer, since it works differently.
     The transformer remembers all previous states and attends to them.
 
     :param src_mask: mask for source inputs, 0 for positions after </s>
-    :param embed: target embedding layer
-    :param bos_index: index of <s> in the vocabulary
-    :param eos_index: index of </s> in the vocabulary
     :param max_output_length: maximum length for the hypotheses
-    :param decoder: decoder to use for greedy decoding
+    :param model: model to use for greedy decoding
     :param encoder_output: encoder hidden states for attention
     :param encoder_hidden: encoder final state (unused in Transformer)
     :return:
         - stacked_output: output hypotheses (2d array of indices),
         - stacked_attention_scores: attention scores (3d array)
     """
-
+    bos_index = model.bos_index
+    eos_index = model.eos_index
     batch_size = src_mask.size(0)
 
     # start with BOS-symbol for each sentence in the batch
@@ -135,22 +126,23 @@ def transformer_greedy(
 
     # a subsequent mask is intersected with this in decoder forward pass
     trg_mask = src_mask.new_ones([1, 1, 1])
+    if isinstance(model, torch.nn.DataParallel):
+        trg_mask = torch.stack(
+            [src_mask.new_ones([1, 1]) for _ in model.device_ids])
 
-    finished = src_mask.new_zeros((batch_size)).byte()
+    finished = src_mask.new_zeros(batch_size).byte()
 
     for _ in range(max_output_length):
-
-        trg_embed = embed(ys)  # embed the previous tokens
-
         # pylint: disable=unused-variable
         with torch.no_grad():
-            logits, out, _, _ = decoder(
-                trg_embed=trg_embed,
+            logits, _, _, _ = model(
+                return_type="decode",
+                trg_input=ys, # model.trg_embed(ys) # embed the previous tokens
                 encoder_output=encoder_output,
                 encoder_hidden=None,
                 src_mask=src_mask,
                 unroll_steps=None,
-                hidden=None,
+                decoder_hidden=None,
                 trg_mask=trg_mask
             )
 
@@ -171,30 +163,23 @@ def transformer_greedy(
 
 
 # pylint: disable=too-many-statements,too-many-branches
-def beam_search(
-        decoder: Decoder,
-        size: int,
-        bos_index: int, eos_index: int, pad_index: int,
-        encoder_output: Tensor, encoder_hidden: Tensor,
-        src_mask: Tensor, max_output_length: int, alpha: float,
-        embed: Embeddings, n_best: int = 1) -> (np.array, np.array):
+def beam_search(model: Model, size: int,
+                encoder_output: Tensor, encoder_hidden: Tensor,
+                src_mask: Tensor, max_output_length: int,
+                alpha: float, n_best: int = 1) -> (np.array, np.array):
     """
     Beam search with size k.
     Inspired by OpenNMT-py, adapted for Transformer.
 
     In each decoding step, find the k most likely partial hypotheses.
 
-    :param decoder:
+    :param model:
     :param size: size of the beam
-    :param bos_index:
-    :param eos_index:
-    :param pad_index:
     :param encoder_output:
     :param encoder_hidden:
     :param src_mask:
     :param max_output_length:
     :param alpha: `alpha` factor for length penalty
-    :param embed:
     :param n_best: return this many hypotheses, <= beam (currently only 1)
     :return:
         - stacked_output: output hypotheses (2d array of indices),
@@ -204,14 +189,19 @@ def beam_search(
     assert n_best <= size, 'Can only return {} best hypotheses.'.format(size)
 
     # init
-    transformer = isinstance(decoder, TransformerDecoder)
+    bos_index = model.bos_index
+    eos_index = model.eos_index
+    pad_index = model.pad_index
+    trg_vocab_size = model.decoder.output_size
+    device = encoder_output.device
+    transformer = isinstance(model.decoder, TransformerDecoder)
     batch_size = src_mask.size(0)
     att_vectors = None  # not used for Transformer
 
     # Recurrent models only: initialize RNN hidden state
     # pylint: disable=protected-access
     if not transformer:
-        hidden = decoder._init_hidden(encoder_hidden)
+        hidden = model.decoder._init_hidden(encoder_hidden)
     else:
         hidden = None
 
@@ -226,12 +216,14 @@ def beam_search(
     # Transformer only: create target mask
     if transformer:
         trg_mask = src_mask.new_ones([1, 1, 1])  # transformer only
+        if isinstance(model, torch.nn.DataParallel):
+            trg_mask = torch.stack(
+                [src_mask.new_ones([1, 1]) for _ in model.device_ids])
     else:
         trg_mask = None
 
     # numbering elements in the batch
-    batch_offset = torch.arange(
-        batch_size, dtype=torch.long, device=encoder_output.device)
+    batch_offset = torch.arange(batch_size, dtype=torch.long, device=device)
 
     # numbering elements in the extended batch, i.e. beam size copies of each
     # batch element
@@ -240,7 +232,7 @@ def beam_search(
         batch_size * size,
         step=size,
         dtype=torch.long,
-        device=encoder_output.device)
+        device=device)
 
     # keeps track of the top beam size hypotheses to expand for each element
     # in the batch to be further decoded (that are still "alive")
@@ -248,10 +240,10 @@ def beam_search(
         [batch_size * size, 1],
         bos_index,
         dtype=torch.long,
-        device=encoder_output.device)
+        device=device)
 
     # Give full probability to the first beam on the first step.
-    topk_log_probs = torch.zeros(batch_size, size, device=encoder_output.device)
+    topk_log_probs = torch.zeros(batch_size, size, device=device)
     topk_log_probs[:, 1:] = float("-inf")
 
     # Structure that holds finished hypotheses.
@@ -264,7 +256,6 @@ def beam_search(
     }
 
     for step in range(max_output_length):
-
         # This decides which part of the predicted sentence we feed to the
         # decoder to make the next prediction.
         # For Transformer, we feed the complete predicted sentence so far.
@@ -278,17 +269,18 @@ def beam_search(
         # decode one single step
         # logits: logits for final softmax
         # pylint: disable=unused-variable
-        trg_embed = embed(decoder_input)
-        logits, hidden, att_scores, att_vectors = decoder(
-            encoder_output=encoder_output,
-            encoder_hidden=encoder_hidden,
-            src_mask=src_mask,
-            trg_embed=trg_embed,
-            hidden=hidden,
-            prev_att_vector=att_vectors,
-            unroll_steps=1,
-            trg_mask=trg_mask  # subsequent mask for Transformer only
-        )
+        with torch.no_grad():
+            logits, hidden, att_scores, att_vectors = model(
+                return_type="decode",
+                encoder_output=encoder_output,
+                encoder_hidden=encoder_hidden,
+                src_mask=src_mask,
+                trg_input=decoder_input, #trg_embed = embed(decoder_input)
+                decoder_hidden=hidden,
+                att_vector=att_vectors,
+                unroll_steps=1,
+                trg_mask=trg_mask  # subsequent mask for Transformer only
+            )
 
         # For the Transformer we made predictions for all time steps up to
         # this point, so we only want to know about the last time step.
@@ -309,7 +301,7 @@ def beam_search(
             curr_scores /= length_penalty
 
         # flatten log_probs into a list of possibilities
-        curr_scores = curr_scores.reshape(-1, size * decoder.output_size)
+        curr_scores = curr_scores.reshape(-1, size * trg_vocab_size)
 
         # pick currently best top k hypotheses (flattened order)
         topk_scores, topk_ids = curr_scores.topk(size, dim=-1)
@@ -321,8 +313,8 @@ def beam_search(
             topk_log_probs = topk_scores.clone()
 
         # reconstruct beam origin and true word ids from flattened order
-        topk_beam_index = topk_ids.floor_divide(decoder.output_size)
-        topk_ids = topk_ids.fmod(decoder.output_size)
+        topk_beam_index = topk_ids.floor_divide(trg_vocab_size)
+        topk_ids = topk_ids.fmod(trg_vocab_size)
 
         # map beam_index to batch_index in the flat representation
         batch_index = (
@@ -418,3 +410,48 @@ def beam_search(
                                        pad_value=pad_index)
 
     return final_outputs, None
+
+
+def run_batch(model: Model, batch: Batch, max_output_length: int,
+              beam_size: int, beam_alpha: float) -> (np.array, np.array):
+    """
+    Get outputs and attentions scores for a given batch
+
+    :param model: Model class
+    :param batch: batch to generate hypotheses for
+    :param max_output_length: maximum length of hypotheses
+    :param beam_size: size of the beam for beam search, if 0 use greedy
+    :param beam_alpha: alpha value for beam search
+    :return: stacked_output: hypotheses for batch,
+        stacked_attention_scores: attention scores for batch
+    """
+    with torch.no_grad():
+        encoder_output, encoder_hidden, _, _ = model(
+            return_type="encode", src=batch.src,
+            src_length=batch.src_length,
+            src_mask=batch.src_mask)
+
+    # if maximum output length is not globally specified, adapt to src len
+    if max_output_length is None:
+        max_output_length = int(max(batch.src_length.cpu().numpy()) * 1.5)
+
+    # greedy decoding
+    if beam_size < 2:
+        stacked_output, stacked_attention_scores = greedy(
+            src_mask=batch.src_mask,
+            max_output_length=max_output_length,
+            model=model,
+            encoder_output=encoder_output,
+            encoder_hidden=encoder_hidden)
+        # batch, time, max_src_length
+    else:  # beam search
+        stacked_output, stacked_attention_scores = beam_search(
+            model=model,
+            size=beam_size,
+            encoder_output=encoder_output,
+            encoder_hidden=encoder_hidden,
+            src_mask=batch.src_mask,
+            max_output_length=max_output_length,
+            alpha=beam_alpha)
+
+    return stacked_output, stacked_attention_scores
